extern crate rand;
extern crate serialize;
use std::{mem};

#[deriving(PartialEq, Eq, Clone, Show)]
enum Brick {
  Brick22 = 0,
  Brick1 = 1,
  Brick2 = 2,
  Brick3 = 3,
  Brick4 = 4,
}

static BRICK_COUNT: uint = 5;
static BRICK_COLORS: uint = 5;

#[deriving(PartialEq, Eq, Clone, Show)]
struct PlacedBrick {
  x: uint,
  y: uint,
  brick: Brick,
  rotated: bool,
}

#[deriving(PartialEq, Eq, Clone, Show)]
struct Floor {
  layout: Vec<PlacedBrick>,
  brick_counts: Vec<uint>,
  fill: Vec<bool>,
}

#[deriving(Decodable, Show)]
struct TowerSpec {
  width: uint,
  height: uint,
  full: bool,
  colored: bool,
  bricks: Vec<BrickSpec>,
}

#[deriving(Decodable, Show)]
struct BrickSpec {
  color: uint,
  brick: uint,
  count: uint,
}

#[deriving(Show)]
struct BuildPlan {
  bricks: Vec<Vec<PlacedBrick>>,
}

fn random_floor<R: rand::Rng>(rng: &mut R, w: uint, h: uint, max_bricks: &[uint], hollow: bool) -> Option<Floor> {
  let mut fill = Vec::from_fn(w*h, |_| false);
  let mut layout = Vec::new();
  let mut brick_count = Vec::from_fn(BRICK_COUNT, |_| 0);
  let mut avail_bricks = max_bricks.to_owned();
  let mut avail_brick_count = max_bricks.iter().fold(0, |a, &b| a + b);

  for x in range(0, w) {
    for y in range(0, h) {
      if *fill.get(y*w + x) { continue }
      if avail_brick_count <= 0 { return None }

      /*
      if hollow {
        if x != 0 && y != 0 && x != w-1 && y != h-1 {
          if rng.gen_weighted_bool(2) { continue }
        }
      }
      */

      let mut placed_brick = None;
      for _ in range(0, 20) {
        let brick = match gen_brick(rng, avail_bricks.as_slice()) {
          Some(b) => b,
          None => return None,
        };

        let rot = 
          if hollow && (x == 0 || x == w - 1) {
            rng.gen_weighted_bool(10)
          } else if hollow && (y == 0 || y == h - 1) {
            !rng.gen_weighted_bool(10)
          } else {
            rng.gen_weighted_bool(2)
          };

        let pl_brick = PlacedBrick {x:x, y:y, brick:brick, rotated:rot };
        if can_be_placed(fill.as_slice(), w, h, pl_brick) {
          placed_brick = Some(pl_brick);
          break;
        }
      }

      match placed_brick {
        Some(brick) => {
          *brick_count.get_mut(brick.brick as uint) += 1;
          *avail_bricks.get_mut(brick.brick as uint) -= 1;
          avail_brick_count -= 1;
          fill_brick(fill.as_mut_slice(), w, h, brick);
          layout.push(brick);
        },
        None => { return None },
      }
    }
  }

  return Some(Floor { layout:layout, brick_counts:brick_count, fill:fill });

  fn gen_brick<R: rand::Rng>(rng: &mut R, avail_bricks: &[uint]) -> Option<Brick> {
    for _ in range(0, 20) {
      let brick_no = rng.gen_range(0, BRICK_COUNT);
      if avail_bricks[brick_no] > 0 {
        return Some(unsafe { mem::transmute(brick_no as u8) });
      }
    }
    None
  }

  fn can_be_placed(fill: &[bool], w: uint, h: uint, brick: PlacedBrick) -> bool {
    let (brick_w, brick_h) = brick_dim(&brick);
    if brick.x + brick_w > w || brick.y + brick_h > h { return false };
    for x in range(brick.x, brick.x + brick_w) {
      for y in range(brick.y, brick.y + brick_h) {
        if fill[y*w + x] { return false }
      }
    }
    true
  }

  fn fill_brick(fill: &mut [bool], w: uint, _h: uint, brick: PlacedBrick) {
    let (brick_w, brick_h) = brick_dim(&brick);
    for x in range(brick.x, brick.x + brick_w) {
      for y in range(brick.y, brick.y + brick_h) {
        assert!(!fill[y*w + x]);
        fill[y*w + x] = true;
      }
    }
  }

}

fn brick_dim(brick: &PlacedBrick) -> (uint, uint) {
  match (brick.brick, brick.rotated) {
    (Brick22, _) => (2, 2),
    (Brick1, _) => (1, 1),
    (Brick2, true) => (2, 1),
    (Brick2, false) => (1, 2),
    (Brick3, true) => (3, 1),
    (Brick3, false) => (1, 3),
    (Brick4, true) => (4, 1),
    (Brick4, false) => (1, 4),
  }
}

fn build_tower(max_bricks: &[uint], w: uint, h: uint, floors: &[Floor], hollow: bool) -> BuildPlan {
  let mut avail_bricks = max_bricks.to_owned();
  let mut plan = BuildPlan { bricks: Vec::new() };
  let mut last_fill = &Vec::from_fn(w*h, |_| true);
  let mut last_full = 0;

  for f in range(0, floors.len()) {
    let floor = &floors[f];
    let mut can_be_added = true;;

    for b in range(0, BRICK_COUNT) {
      if floor.brick_counts.get(b) > avail_bricks.get(b) {
        can_be_added = false;
        break;
      }
    }

    if hollow && can_be_added {
      for pl_brick in floor.layout.iter() {
        let (brick_w, brick_h) = brick_dim(pl_brick);
        let mut is_supported = false;
        for x in range(pl_brick.x, pl_brick.x + brick_w) {
          for y in range(pl_brick.y, pl_brick.y + brick_h) {
            is_supported = is_supported || *last_fill.get(y*w + x);
          }
        }

        can_be_added = can_be_added && is_supported;
      }
    }

    if hollow {
      let mut is_full = true;
      for i in range(0, w*h) {
        is_full = is_full && *floor.fill.get(i);
      }

      if is_full {
        last_full = plan.bricks.len() + 1;
      }
    }

    if can_be_added {
      for b in range(0, BRICK_COUNT) {
        *avail_bricks.get_mut(b) -= *floor.brick_counts.get(b);
      }
      last_fill = &floor.fill;
      plan.bricks.push(floor.layout.clone());
    }
  }

  if hollow {
    plan.bricks.truncate(last_full);
  }
  plan
}


fn main() {
  use rand::{Rng};
  use serialize::{json, Decodable};
  use std::{str, io};

  let input_bytes = io::stdin().read_to_end().unwrap();
  let input_str = str::from_utf8(input_bytes.as_slice()).unwrap();
  let json_obj = json::from_str(input_str).unwrap();
  let mut json_decoder = json::Decoder::new(json_obj);
  let tower_spec: TowerSpec = Decodable::decode(&mut json_decoder).unwrap();

  if !tower_spec.colored {
    let hollow = !tower_spec.full;
    let w = tower_spec.width;
    let h = tower_spec.height;

    let mut max_bricks = Vec::from_fn(BRICK_COUNT, |_| 0);
    let mut brick_colors = Vec::from_fn(BRICK_COUNT * BRICK_COLORS, |_| 0);
    for brick_spec in tower_spec.bricks.iter() {
      *max_bricks.get_mut(brick_spec.brick) += brick_spec.count;
      *brick_colors.get_mut(brick_spec.brick*BRICK_COLORS + brick_spec.color) += brick_spec.count;
    }

    let mut rng = rand::XorShiftRng::new_unseeded();
    let mut floors: Vec<Floor> = range(0, 2_000)
      .filter_map(|_| random_floor(&mut rng, w, h, max_bricks.as_slice(), hollow))
      .collect();
    floors.sort_by(|f1, f2| {
      let brs1 = f1.brick_counts.iter().fold(0, |a, &b| a + b);
      let brs2 = f2.brick_counts.iter().fold(0, |a, &b| a + b);
      brs1.cmp(&brs2)
    });
    floors.truncate(1_000);

    let max_plan = range(0, 10_000).map(|_| {
      rng.shuffle(floors.as_mut_slice());
      build_tower(max_bricks.as_slice(), w, h, floors.as_slice(), hollow)
    }).max_by(|plan| plan.bricks.len())
      .unwrap_or(BuildPlan { bricks: Vec::new() });

    let tower_type_str = match (tower_spec.full, tower_spec.colored) {
      (false, false) => "D",
      (true, false) => "P",
      (false, true) => "DB",
      (true, true) => "PB",
    };

    println!("{} {} {}", w, h, tower_type_str);
    let mut avail_colors = brick_colors.clone();
    for f in range(0, max_plan.bricks.len()) {
      let bricks = max_plan.bricks.get(f);
      for brick in bricks.iter() {
        let brick_ty = brick.brick as uint;
        let mut color = None;
        for c in range(0, BRICK_COLORS) {
          if *avail_colors.get(brick_ty*BRICK_COLORS + c) > 0 {
            *avail_colors.get_mut(brick_ty*BRICK_COLORS + c) -= 1;
            color = Some(c);
            break;
          }
        }

        let color_str = match color.expect("No color") {
          0 => "A", 1 => "R", 2 => "G",
          3 => "B", 4 => "Y", _ => unreachable!(),
        };

        let (dim_x, dim_y) = brick_dim(brick);
        println!("{} {} {} {} {}x{}", f, brick.x, brick.y, color_str, dim_x, dim_y);
      }
    }
    //println!("// {}", max_plan.bricks.len());
  }
}
